#!/usr/global/bin/perl -w

# Stand-alone server process for the abbreviations database
# $Id: ABserver.pl,v 1.5 1998/02/09 20:00:11 bates Exp $

## Copyright (C) 1997, 1998 Douglas M. Bates and Duncan J. Murdoch

## Version 1.5a modified by Duncan Murdoch:  added PIDFile, changed
## abbreviation filenames and descriptions

## This file is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2, or (at your option)
## any later version.

## This file is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this code; see the file COPYING.  If not, write to
## the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

require 5.003;
use strict;
use Carp;
use Socket;
use Search::Dict;
use FileHandle;

# Change these config options to suit your local installation

my @Files = (           # Modify if necessary
             { 'Name' => 'ABBR.V97',
               'Desc' => 'Full-length names',
               'Pre'  => '#'},
             { 'Name' => 'ABB.V97',
               'Desc' => 'Shortened names',
               'Pre'  => '#'},
             { 'Name' => 'ABBRAMS.V97',
               'Desc' => 'Full-length AMS forms',
               'Pre'  => '#'},
             { 'Name' => 'ABAMS.V97',
               'Desc' => 'Shortened AMS forms',
               'Pre'  => '#'},
             { 'Name' => 'AB.V97',
               'Desc' => 'No delimiters',
               'Pre'  => ''}
             );

my %Config = ('Port' => 1695,
              'ABBRDir' => "/home/cised/1997/ABBREV",
              'TimeOut' => 1000, # default timeout is 1000 minutes
              'PIDFile' => "/home/cised/ABcurrent.pid",
              );

my $Usage = "$0 (Configuration_file)";
if (my $Config = shift) {   # read the configuration file if given
    die "$Usage\n" unless -r $Config;
    open(CONFIG, "< $Config");  # Load CONFIG into %Config
    while (<CONFIG>) {
        next if /^#/;       # skip comment lines
        next unless /\w/;   # and blank lines
        chomp;
        die "Configuration line is not of form \"Key = value\"\n$_\n"
            unless (/(\w+)\s*=\s*(.*)$/);
        $Config{$1} = ($Config{$1}) ? $Config{$1} . "\n" . $2 : $2;
    }
}

chdir($Config{'ABBRDir'}) ||
    die "Can't change directory to $Config{'ABBRDir'}: $!\n";

if ($Config{'PIDFile'}) {
    if (open(PID, "> $Config{'PIDFile'}")) {
        print PID $$;
        close(PID);
    } else {
        warn "unable to open PIDFile $Config{'PIDFile'}: $!";
    }
}

my $Port = $Config{'Port'};
if ($Port =~ /\D/) { $Port = getservbyname($Port, 'tcp') }
my $proto = getprotobyname('tcp');
socket(S, PF_INET, SOCK_STREAM, $proto)               || die "socket: $!";
setsockopt(S, SOL_SOCKET, SO_REUSEADDR, pack("l", 1)) || die "setsockopt: $1";
bind(S, sockaddr_in($Port, INADDR_ANY))               || die "bind: $!";
listen(S, SOMAXCONN)                                  || die "connect: $!";

select(NS); $| = 1; select(STDOUT);

my $uid;
if ($Config{'User'}) {      # attempt to change the effective uid
    warn "Username $Config{'User'} is not valid\n"
        unless $uid = (getpwnam($Config{'User'}))[2];
}
($<,$>) = ($uid, $uid) if $uid;

select(S); $| = 1; select(stdout);
for(;;) {
    (my $addr = accept(NS,S)) || die $!;
    my $child;
    if (($child = fork()) == 0) {
        if ((my $grandchild = fork()) == 0) {
            my ($Port,$inetaddr) = sockaddr_in($addr);
            my $hostname = gethostbyaddr($inetaddr, AF_INET);
            $inetaddr = inet_ntoa($inetaddr);
            if ($Config{'AddressMask'}) { # Check address
                unless ($inetaddr =~ /$Config{'AddressMask'}/) {
                    print NS "010 Connections from address $inetaddr " .
                        "not allowed on this server\n";
                    close NS;
                    exit;
                }
            }

            print NS "099 Connected to server - please send login name\n";
            my $logname = <NS>; chop($logname);
            chop($logname) if $logname =~ /\r$/;
            for (0..$#Files) {  # select from available abbreviation files
                print NS $_, " ", $Files[$_]{'Desc'}, " (", $Files[$_]{'Name'},
                ")\n";
            }
            print NS "098 Please select an abbreviation file by number \n";
            my $FileNo = <NS>; chop($FileNo); $FileNo += 0;
            $FileNo = 0 if ($FileNo < 0 || $FileNo > $#Files);

            my $AbbrFile = new FileHandle "< $Files[$FileNo]{'Name'}"
                || croak "open($Files[$FileNo]{'Name'}: $!\n";
            my $Pre = $Files[$FileNo]{'Pre'};

            my $rin = '';
            vec($rin, fileno(NS), 1) = 1; my $ein = $rin;
            while (1) {
                print NS "100 Ready\n";
                my ($nfound, $timeleft) = select(my $rout=$rin, undef,
                                                 my $eout=$ein, my $TimeOut);
                if ($nfound && vec($rout, fileno(NS), 1)) { # can read a line
                    $_ = <NS>;
                    chop; chop if /\r$/; # For testing with telnet.
                    last if /^q$/;
                    &look($AbbrFile, $Pre . $_, 0, 1);
                    my $Expansion = <$AbbrFile>;
                    if ( $Expansion =~ /$Pre$_/ ) {
                        print NS "210 ", $Expansion;
                    } else {
                        print NS "220 ", $_, "\n";
                    }
                } elsif ($nfound && vec($eout, fileno(NS), 1)) { # exception
                    print NS "102 Exception on socket\n";
                    close NS;
                    exit;
                } else {    # Timed out
                    print NS "103 Timing Out after ",$TimeOut/60," minutes\n";
                    close NS;
                    exit;
                }
            }
            print NS "101 Goodbye\n";
            close NS;
            exit;
        }
        exit;           # child exits after forking grandchild
    }
    waitpid($child, 0);
    close NS;
}




